"""
w2lapp.addmodifyform.py: input form for adding and modifying an entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: addmodifyform.py,v 1.97 2003/06/01 11:32:31 michael Exp $
"""

import msbase,ldap.schema,ldaputil.base,\
       w2lapp.core,w2lapp.cnf,w2lapp.form,w2lapp.gui
from ldaputil.base import ParentDN
from msbase import listremove

heading_msg = {'modifyform':'Modify entry','addform':'Add new entry'}


class InputFormEntry(ldap.schema.Entry):

  def __init__(self,sid,form,ls,dn,schema,entry):
    assert type(dn)==type(u'')
    ldap.schema.Entry.__init__(self,schema,None,entry)
    self.sid = sid
    self.form = form
    self.ls = ls
    self.set_dn(dn)

  def set_dn(self,dn):
    self.dn = dn
    self.rdn_dict = self._get_rdn_dict(dn)

  def _get_rdn_dict(self,dn):
    assert type(dn)==type(u'')
    rdn_dict = ldap.schema.Entry(self._s,None,ldaputil.base.rdn_dict(dn,self.ls.charset))
    for attr_type,attr_values in rdn_dict.items():
      del rdn_dict[attr_type]
      d = ldap.cidict.cidict()
      for attr_value in attr_values:
        assert type(attr_value)==type(u'')
        d[attr_value] = None
      rdn_dict[attr_type] = d
    return rdn_dict

  def __getitem__(self,nameoroid):
    oid = self._at2key(nameoroid)[0]
    try:
      attr_values = ldap.schema.Entry.__getitem__(self,nameoroid)
    except KeyError:
      attr_values = []
    else:
      attr_values.sort()
    attr_values = attr_values or ['']
    result = []
    nameoroid_se = self._s.get_obj(ldap.schema.AttributeType,oid)
    for attr_value in attr_values:
      attr_value = unicode(attr_value,self.ls.charset,'replace')
      attr_value_disp = w2lapp.core.utf2display(self.form.accept_charset,attr_value)
      if nameoroid.lower()=='objectclass' or \
         (self.rdn_dict.has_key(nameoroid) and self.rdn_dict[nameoroid].has_key(attr_value)):
        input_field_str = w2lapp.gui.HIDDEN_FIELD % (
          'in_value',attr_value_disp,attr_value_disp
        )
      else:
        syntax_class = w2lapp.schema.syntaxes.oid2syntax.get(
          w2lapp.schema.syntaxes.syntax_registry.get(oid,nameoroid_se.syntax),
          w2lapp.schema.syntaxes.LDAPSyntax
        )
        input_field = syntax_class(nameoroid,attr_value).formField()
        input_field.charset = self.form.accept_charset
        input_field.name = 'in_value'
        input_field_str = input_field.inputHTML()
      result.append('\n'.join([
        w2lapp.gui.HIDDEN_FIELD % ('in_attrtype',nameoroid,''),
        input_field_str
      ]))
    return '<br>\n'.join(result)


def PrintInputFieldset(
  outf,sid,form,ls,dn,sub_schema,attr_types_dict,fieldset_title,entry
):
  outf.write("""<fieldset title="%s">
    <legend>%s</legend>
    <dl>
    """ % (fieldset_title,fieldset_title)
  )
  attr_type_names = ldap.cidict.cidict()
  for a in entry.keys():
    if attr_types_dict.has_key(entry._at2key(a)[0]):
      attr_type_names[a] = None
  for a in attr_types_dict.values():
    if w2lapp.schema.editable_attr(sub_schema,a.oid):
      attr_type_names[(a.names or (a.oid,))[0]] = None
  attr_types = attr_type_names.keys()
  attr_types.sort()
  for attr_type in attr_types:
    outf.write('<dt>%s</dt>\n' % (w2lapp.gui.SchemaElementName(sid,form,dn,sub_schema,attr_type,ldap.schema.AttributeType)))
    outf.write('<dd>\n%s\n</dd>' % (entry[attr_type]))
  outf.write('</dl></fieldset>')
  return # PrintInputFieldset()


def ObjectClassForm(
  sid,outf,form,command,ls,sub_schema,dn,rdn,all_oc
):
  """Form for choosing object class(es)"""

  existing_object_classes = ls.getObjectClasses(dn)

  command_hidden_fields = [('dn',dn)]

  if command=='addform' and rdn:
    command_hidden_fields.append(('add_rdn',rdn))

  if command=='addform':
    existing_structural_oc = []
    existing_auxiliary_oc = []
  else:
    existing_structural_oc,dummy,existing_auxiliary_oc = w2lapp.schema.object_class_categories(sub_schema,existing_object_classes)

  all_structural_oc,dummy,all_auxiliary_oc = w2lapp.schema.object_class_categories(sub_schema,all_oc)

  structural_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',options=all_structural_oc,default=existing_structural_oc
  )
  auxiliary_select_field = w2lapp.form.ObjectClassSelect(
    name='ldap_oc',options=all_auxiliary_oc,default=existing_auxiliary_oc
  )

  Msg = {
    'addform':'Choose object class(es) for new entry or choose from quick-list.',
    'modifyform':'You may change the object class(es) for the entry.',
  }[command]

  # Write HTML output of object class input form
  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,
    'Object Class Selection',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

  outf.write("""
    <div id="Input" class="Main">\n<h1>%s</h1>
    %s<input type="submit" value="Input form &gt;&gt;">
    %s
    <p>%s (%s)</p>
    <table summary="Choosing object classes">
    <tr>
      <td>Structural object classes<br>%s</td>
      <td>Auxiliary object classes<br>%s</td>
  """ % (
    heading_msg[command],
    form.beginFormHTML(command,sid,'GET',None),
    ''.join([
      w2lapp.gui.HIDDEN_FIELD % (param_name,w2lapp.core.utf2display(form.accept_charset,param_value,escape_entities=1),'')
      for param_name,param_value in command_hidden_fields
    ]),
    Msg,
    form.applAnchor('oid','Browse Schema',sid,[('dn',dn)]),
    structural_select_field.inputHTML(None),
    auxiliary_select_field.inputHTML(None),
  ))

  # Select list with all configured object classes
  if command=='addform':
    outf.write('<td>\n')
    default_oc_list = w2lapp.cnf.GetParam(ls,'addform_oc_list',[])
    for desc,rdnAttrType,oc_list in default_oc_list:
      quick_button_command_hidden_fields = [
        ('dn',dn),('add_rdn','%s=' % (rdnAttrType.encode()))
      ]
      quick_button_command_hidden_fields.extend(
        [ ('ldap_oc',oc) for oc in oc_list ]
      )
      outf.write('<p>\n%s | %s\n</p>\n' % (
        form.applAnchor(command,desc,sid,quick_button_command_hidden_fields),
        ', '.join(oc_list)
      ))
    outf.write('</td>\n')

  outf.write('</tr>\n</table>\n</form>\n</div>')
  w2lapp.gui.PrintFooter(outf,form)
  return # ObjectClassForm()


def read_old_entry(ls,dn,schema):
  # Retrieve all attribute types of the entry first
  # Build a list of attributes to be read
  read_attrs = msbase.intersection(
    w2lapp.cnf.GetParam(ls,'requested_attrs',[]),
    [
      a
      for a in ls.getAttributeTypes(dn,None,opAttrs=0)
      if w2lapp.schema.editable_attr(schema,a)
    ],
    ignorecase=1
  )
  # Read the editable attribute valus of entry
  return ldap.schema.Entry(schema,dn,ls.readEntry(dn,read_attrs,no_cache=1)[0][1])


def read_old_entry(ls,dn,schema):
  # Retrieve all attribute types of the entry first
  entry_attr_types = ls.getAttributeTypes(
    dn,None,
    opAttrs=0
  )
  # Build a list of attributes to be read
  read_attrs = ldap.cidict.cidict()
  for a in w2lapp.cnf.GetParam(ls,'requested_attrs',[]):
    read_attrs[a] = None
  for a in entry_attr_types:
    if w2lapp.schema.editable_attr(schema,a):
      read_attrs[a] = None
  # Read the editable attribute valus of entry
  return ldap.schema.Entry(schema,dn,ls.readEntry(dn,read_attrs.keys(),no_cache=1)[0][1])


def w2l_AddModifyForm(sid,outf,command,form,ls,dn,Msg='',rdn_default='',entry=None,skip_oc_input=0):

  ObjectClass = ldap.schema.ObjectClass

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  ################################################################
  # Determine entry
  ################################################################

  if not entry and (command=='modifyform'):
    entry = read_old_entry(ls,dn,sub_schema)
  else:
    entry = ldap.schema.Entry(sub_schema,dn.encode(ls.charset),entry or {})

  ################################################################
  # Build a list of all available object classes
  ################################################################

  available_object_classes = [
    (sub_schema.get_obj(ObjectClass,oid).names or (oid,))[0]
    for oid in sub_schema.listall(ObjectClass)
  ]
  available_object_classes.sort()

  ################################################################
  # Handle ldap_oc input
  ################################################################

  if 'ldap_oc' in form.inputFieldNames:
    # Read objectclass(es) from input form
    entry['objectClass'] = [ oc.encode('ascii') for oc in form.field['ldap_oc'].value ]

  elif not skip_oc_input:

    ##############################################################
    # User did not specify ldap_oc
    # => output a form for selecting object class(es)
    # Determine which RDN should be printed as input
    ##############################################################

    # Output the web page with object class input form
    ObjectClassForm(
      sid,outf,form,command,ls,sub_schema,dn,
      form.getInputValue('add_rdn',[rdn_default])[0],
      available_object_classes      
    )
    return

  ################################################################
  # Check whether there are object classes requested which
  # are not existent in current subschema
  ################################################################

  notavail_object_classes = msbase.listremove(
    entry['objectClass'],available_object_classes,ignorecase=1
  )
  if notavail_object_classes:
    Msg = Msg+'<p><strong>Warning!</strong>: Following object class(es) not available in subschema: %s</p>' % (
      ', '.join(notavail_object_classes)
    )

  ################################################################

  required_attrs_dict,allowed_attrs_dict = entry.attribute_types(
    attr_type_filter = [
      ('no_user_mod',[0]),
      ('usage',range(2)),
    ],
    strict=0,raise_keyerror=0
  )

  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,
    heading_msg[command],
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

  outf.write("""<div id="Input" class="Main">
    <h1>%s</h1>
    %s
    %s
    <input type="hidden" name="dn" value="%s">
    <input type="submit" value="%s">
    <input type="reset" value="Reset">
    <br>
    """ % (
      heading_msg[command],
      Msg,
      form.beginFormHTML(command[:-4],sid,'POST',None),
      dn.encode(form.accept_charset),
      '%s%s' % (command[0].upper(),command[1:-4])
    )
  )

  if command=='addform':
    f = form.field['add_rdn']
    if 'add_rdn' in form.inputFieldNames:
      f.setDefault(f.value[0])
    elif len(required_attrs_dict)<=2:
      for at_oid in required_attrs_dict.keys():
        if at_oid!='2.5.4.0':
          rdn_at_name = (required_attrs_dict[at_oid].names or (at_oid,))[0]
          f.setDefault(rdn_at_name+'=')
    outf.write('<p>RDN: %s</p>' % f.inputHTML())

  input_form_entry = InputFormEntry(sid,form,ls,dn,sub_schema,entry)

  PrintInputFieldset(outf,sid,form,ls,dn,sub_schema,required_attrs_dict,'Required attributes',input_form_entry)
  PrintInputFieldset(outf,sid,form,ls,dn,sub_schema,allowed_attrs_dict,'Allowed attributes',input_form_entry)

  if command=='modifyform':
    nomatching_attrs_dict = ldap.cidict.cidict()
    for at_name in entry.keys():
      # Write hidden form parameters with attribute type names
      # of all existing attributes
      outf.write(w2lapp.gui.HIDDEN_FIELD % ('in_oldattrtypes',at_name,''))
      # Output attributes which does not appear in the schema but
      # do exist in the entry
      try:
        at_oid = sub_schema.name2oid[ldap.schema.AttributeType][at_name]
      except KeyError:
        pass
      else:
        if not (
          allowed_attrs_dict.has_key(at_oid) or \
          required_attrs_dict.has_key(at_oid) or \
          at_name.lower()=='objectclass'
        ):
          nomatching_attrs_dict[at_oid] = sub_schema.get_obj(ldap.schema.AttributeType,at_oid)
    if nomatching_attrs_dict:
      PrintInputFieldset(outf,sid,form,ls,dn,sub_schema,nomatching_attrs_dict,'Existing attributes not matching schema',input_form_entry)

  all_attrtypes = [
    (a.names or (a.oid,))[0]
    for a in required_attrs_dict.values()+allowed_attrs_dict.values()
    if w2lapp.schema.editable_attr(sub_schema,a.oid) and not a.single_value
  ]
  all_attrtypes.sort()
  # Display select field
  outf.write("""
    <fieldset>
      <legend>Additional attributes</legend>
      <table>
        <tr><td>
          <select name="in_attrtype"><option>%s</option></select>
          <input name="in_value" size="40" value="">
        </td></tr>
        <tr><td>
          <input name="in_attrtype" size="15" maxsize="200" value="">
          <input name="in_value" size="40" value="">
        </td></tr>
      <table>
    </fieldset>
    <fieldset>
      <legend>Raw LDIF data</legend>
      <textarea name="in_ldif" value="" wrap=off rows="15" cols="75"></textarea>
      <p>
        Notes:
      </p>
      <ul>
        <li>Lines containing "dn:" will be ignored</li>
        <li>Only the first entry (until first empty line) will be accepted</li>
        <li>Maximum length is set to %d bytes</li>
        <li>Allowed URL schemes: %s</li>
      </ul>
    </fieldset>
</form>
</div>
""" % ('</option>\n<option>'.join(all_attrtypes),w2lapp.cnf.misc.ldif_maxbytes,', '.join(w2lapp.cnf.misc.ldif_url_schemes)))

  w2lapp.gui.PrintFooter(outf,form)

