"""
w2lapp.form.py: class for web2ldap input form handling

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: form.py,v 1.16 2003/06/01 13:58:12 michael Exp $
"""

import urllib,re,ldif,ldap.schema,msgzip,ldaputil.base,pyweblib.forms,w2lapp.cnf

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO


class Web2LDAPForm(pyweblib.forms.Form):

  def __init__(self,inf,env,script_name = None):
    pyweblib.forms.Form.__init__(self,inf,env)
    self.script_name = script_name or self.script_name
    self.determineCharset()
    self.initializeForm()

  def outFileObject(self,outf):

    """
    Create file object for gziping on-the-fly if configuration
    and HTTP header allows it
    """
#    outfile = msgzip.DebugFile(outf)
    outfile = outf
    if w2lapp.cnf.misc.gzip_level>0 and \
       not msgzip.GzipFile is None and \
       self.http_accept_encoding['gzip']>0:
      # Create new gzip-stream file object
      return msgzip.GzipFile(
        filename='web2ldap stream output',
        mode='wb',
        compresslevel=w2lapp.cnf.misc.gzip_level,
        fileobj=outfile
      )
    else:
      return outfile

  def determineCharset(self):
    if w2lapp.cnf.misc.print_rawutf8==2:
      self.accept_charset = 'utf-8'
    elif w2lapp.cnf.misc.print_rawutf8==1:
      if self.http_accept_charset['utf-8']:
        self.accept_charset = 'utf-8'
      else:
        self.accept_charset = self.http_accept_charset.preferred()
    else:
      self.accept_charset = 'iso-8859-1'
    return # determineCharset()

  def initializeForm(self):
    """
    Add the required fields 
    """
    self.addGeneralFields()
    self.addCommandFields()

  def addGeneralFields(self):
    self.addField(pyweblib.forms.Input('who',u'Bind DN',255,1,u'.*',size=40))
    self.addField(pyweblib.forms.Input('cred',u'with Password',200,1,u'.*',size=15))
    self.addField(pyweblib.forms.Input('ldapurl',u'LDAP Url',1024,1,'[ ]*ldap(|i|s)://.*',size=30))
    self.addField(pyweblib.forms.Input('host',u'Host:Port',255,1,w2lapp.gui.host_pattern,size=30))
    self.addField(DistinguishedNameInput('dn','Distinguished Name'))
    self.addField(pyweblib.forms.Select('scope','Scope',1,options=[('0','Base'),('1','One level'),('2','Sub tree')],default='2'))
    self.addField(DistinguishedNameInput('login_search_root','Login search root'))
    self.addField(pyweblib.forms.Input('login_filterstr',u'Login search filter string',120,1,'.*'))

  def addCommandFields(self):
    pass

  def actionUrlHTML(self,command,sid):
    return '%s/%s%s' % (
      self.script_name,
      command,
      {0:'/%s' % sid,1:''}[sid is None],
    )

  def beginFormHTML(self,command,sid,method,target=None):
    target = {0:'target="%s"' % (target),1:''}[target is None]
    return """
      <form
        action="%s"
        method="%s"
        %s
        enctype="application/x-www-form-urlencoded"
        accept-charset="%s"
      >
      """  % (self.actionUrlHTML(command,sid),method,target,self.accept_charset)

  def formHTML(
    self,command,submitstr,sid,method,form_parameters,
    extrastr='',
    target=None
  ):
    """
    Build the HTML text of a submit form
    """
    utf2display = w2lapp.core.utf2display
    HIDDEN_FIELD = w2lapp.gui.HIDDEN_FIELD
    form_str = [
      self.beginFormHTML(command,sid,method,target)
    ]
    for param_name,param_value in form_parameters:
      form_str.append(HIDDEN_FIELD % (
          param_name,
          utf2display(self.accept_charset,param_value,escape_entities=1),
          ''
        )
      )
    form_str.append("""
      <input type="submit" value="%s">
      %s
      </form>""" % (submitstr,extrastr)
    )
    return '\n'.join(form_str)

  def applAnchor(
    self,command,anchor_text,sid,
    form_parameters,
    target=None,
  ):
    """
    Build the HTML text of a anchor with form parameters
    """
    charset = self.accept_charset
    utf2display = w2lapp.core.utf2display
    target = {0:'target="%s"' % (target),1:''}[target is None]
    return '<a class="CommandLink" %s href="%s?%s">%s</a>\n' % (
      target,
      self.actionUrlHTML(command,sid),
      '&amp;'.join([
        '%s=%s' % (
          param_name,
          urllib.quote_plus(utf2display(charset,param_value,escape_entities=0))
        )
        for param_name,param_value in form_parameters
      ]),
      anchor_text
    )


class Web2LDAPNullForm(Web2LDAPForm):

  def initializeForm(self):
    pass


class Web2LDAPForm_searchform(Web2LDAPForm):

  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('searchform_mode',u'Search form mode',1,options=[(u'base',u'Base'),(u'adv',u'Advanced'),(u'exp',u'Expert')],default=u'base'))
    self.addField(pyweblib.forms.Field('search_params',u'Anzahl Suchangaben',2,1,'[0-9]+'))
    self.addField(DistinguishedNameInput('search_root','Search root'))
    self.addField(pyweblib.forms.Input('search_filterstr',u'Search filter string',120,1,'.*'))
    self.addField(
      pyweblib.forms.Select(
        'search_resnumber',
        u'Number of results to display',
        1,
        options=[(u'0',u'unlimited'),(u'10',u'10'),(u'20',u'20'),(u'50',u'50'),(u'100',u'100'),(u'200',u'200')],
        default='10'
      )
    )
    self.addField(pyweblib.forms.Input('search_attrs',u'Attributes to be read',250,1,ur'[\w,_;-]+'))

class Web2LDAPForm_search(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('search_filterstr',u'Search filter string',120,1,u'.*'))
    self.addField(pyweblib.forms.Input('search_resminindex',u'Minimum index of search results',10,1,u'[0-9]+'))
    self.addField(pyweblib.forms.Input('search_resnumber',u'Number of results to display',3,1,u'[0-9]+'))
    self.addField(ExportFormatSelect('search_output'))
    self.addField(pyweblib.forms.Select('search_mode',u'Search Mode',1,options=[ur'(&%s)',ur'(|%s)']))
    self.addField(AttributeType('search_attr',u'Search attribute',w2lapp.cnf.misc.max_searchparams))
    self.addField(pyweblib.forms.Select('search_option',u'Search option',w2lapp.cnf.misc.max_searchparams,options=w2lapp.searchform.search_options))
    self.addField(pyweblib.forms.Input('search_string',u'Search string',60,w2lapp.cnf.misc.max_searchparams,u'.*'))
    self.addField(pyweblib.forms.Input('search_attrs',u'Attributes to be read',250,1,ur'[\w,_;-]+'))

class Web2LDAPForm_modifyform(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('ldap_oc',u'Object classes',60,w2lapp.cnf.misc.max_searchparams,u'[a-zA-Z0-9.-]+'))

class Web2LDAPForm_addform(Web2LDAPForm_modifyform):
  def addCommandFields(self):
    Web2LDAPForm_modifyform.addCommandFields(self)
    self.addField(pyweblib.forms.Input('add_rdn','RDN of new entry',255,1,u'.*',size=50))

class Web2LDAPForm_conninfo(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(
      pyweblib.forms.Radio(
        'ldap_managedsait',u'Manage DSA it mode',maxValues=1,
        options=[(u'0',u'disabled'),(u'1',u'enabled')],default=u'0'
      )
    )
    self.addField(
      pyweblib.forms.Select(
        'ldap_deref',u'Dereference aliases',maxValues=1,default=str(ldap.DEREF_NEVER),
        options=[
          (str(ldap.DEREF_NEVER),u'never'),
          (str(ldap.DEREF_SEARCHING),u'searching'),
          (str(ldap.DEREF_FINDING),u'finding'),
          (str(ldap.DEREF_ALWAYS),u'always'),
        ]
      )
    )

class Web2LDAPForm_add(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('add_rdn',u'RDN of new entry',255,1,u'.*',size=50))
    self.addField(AttributeType('in_attrtype',u'Attribute Type',w2lapp.cnf.misc.input_maxattrs))
    self.addField(
      pyweblib.forms.Input(
        'in_value',u'Attribute Value',
        w2lapp.cnf.misc.input_maxfieldlen,w2lapp.cnf.misc.input_maxattrs,
        ('.*',re.U|re.M|re.S)
      )
    )
    self.addField(LDIFTextArea('in_ldif',u'LDIF data'))

class Web2LDAPForm_modify(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(AttributeType('in_attrtype',u'Attribute Type',w2lapp.cnf.misc.input_maxattrs))
    self.addField(AttributeType('in_oldattrtypes',u'Old attribute types',w2lapp.cnf.misc.input_maxattrs))
    self.addField(
      pyweblib.forms.Input(
        'in_value',u'Attribute Value',
        w2lapp.cnf.misc.input_maxfieldlen,w2lapp.cnf.misc.input_maxattrs,
        ('.*',re.U|re.M|re.S)
      )
    )
    self.addField(LDIFTextArea('in_ldif',u'LDIF data'))

class Web2LDAPForm_delete(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('delete_confirm',u'Confirmation',1,options=['yes','no'],default='no'))
    self.addField(pyweblib.forms.Input('delete_attr',u'Attribute to be deleted',255,100,u'[\w_;-]+'))

class Web2LDAPForm_rename(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Input('rename_newrdn',u'New RDN',255,1,ldaputil.base.rdn_pattern,size=50))
    self.addField(DistinguishedNameInput('rename_newsuperior','New Superior DN'))
    self.addField(pyweblib.forms.Checkbox('rename_delold',u'Delete old',1,default="yes",checked=1))

class Web2LDAPForm_passwd(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(DistinguishedNameInput('passwd_who',u'Password DN'))
    self.addField(pyweblib.forms.Field('passwd_newpasswd',u'New password',100,2,'.*'))
    self.addField(pyweblib.forms.Select('passwd_scheme',u'Hash',1,options=w2lapp.passwd.available_hashtypes,default=w2lapp.passwd.available_hashtypes[-1]))
    self.addField(pyweblib.forms.Checkbox('passwd_ntpasswordsync',u'Sync ntPassword for Samba',1,default="yes",checked=1))

class Web2LDAPForm_read(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(pyweblib.forms.Select('read_nocache',u'Force fresh read',1,options=['0','1'],default=0))
    self.addField(pyweblib.forms.Input('read_attr',u'Read attribute',255,100,u'[\w_;-]+'))
    self.addField(pyweblib.forms.Select('read_attrmode',u'Read attribute',1,options=[u'view',u'load']))
    self.addField(pyweblib.forms.Input('read_attrindex',u'Read attribute',255,1,u'[0-9]+'))
    self.addField(pyweblib.forms.Input('read_attrmimetype',u'MIME type',255,1,u'[\w.-]+/[\w.-]+'))
    self.addField(pyweblib.forms.Select('read_output',u'Read output format',1,options=['table','vcard','template'],default='template'))

class Web2LDAPForm_groupadm(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(DistinguishedNameInput('group_search_root','Group search root'))
    self.addField(DistinguishedNameInput('groupadm_add','Add to group',300))
    self.addField(DistinguishedNameInput('groupadm_remove','Remove from group',300))

class Web2LDAPForm_login(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(DistinguishedNameInput('login_who',u'Bind DN'))

class Web2LDAPForm_locate(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(
      pyweblib.forms.Input(
        'locate_name',u'Location name',
        255,1,u'(%s|%s|%s)' % (
          w2lapp.gui.mail_pattern[1:-1],
          w2lapp.gui.host_pattern,
          ldaputil.base.dn_pattern
        ),
        size=25
    ))

class Web2LDAPForm_oid(Web2LDAPForm):
  def addCommandFields(self):
    self.addField(OIDInput('oid',u'OID'))
    self.addField(pyweblib.forms.Select('oid_class','Schema element class',1,options=ldap.schema.SCHEMA_ATTRS,default=''))



FORM_CLASS = {
  '':Web2LDAPForm,
  'monitor':Web2LDAPNullForm,
  'urlredirect':Web2LDAPNullForm,
  'disconnect':Web2LDAPNullForm,
}

_FORM_CLASS_NAME_PREFIX = 'Web2LDAPForm_'
_COMMAND_STR_OFFSET = len(_FORM_CLASS_NAME_PREFIX)

for _name in dir():
  if _name.startswith(_FORM_CLASS_NAME_PREFIX):
    c = eval(_name)
    try:
      command = _name[_COMMAND_STR_OFFSET:]
    except IndexError:
      pass
    else:
      FORM_CLASS[command] = c
      del command



class DistinguishedNameInput(pyweblib.forms.Input):
  """Input field class for LDAP DNs."""

  def __init__(self,name='dn',text='DN',maxValues=1,required=0,default=''):
    pyweblib.forms.Input.__init__(
      self,name,text,1024,maxValues,'',
      size=40,required=required,default=default
    )

  def _validateFormat(self,value):
    if value and not ldaputil.base.is_dn(value):
      raise pyweblib.forms.InvalidValueFormat(
        self.name,
        self.text.encode(self.charset),
        value.encode(self.charset)
      )


class LDIFTextArea(pyweblib.forms.Textarea):
  """A single multi-line input field for LDIF data"""

  def __init__(
    self,name='in_ldif',text='LDIF data',required=0,max_entries=1
  ):
    pyweblib.forms.Textarea.__init__(
      self,
      'in_ldif','LDIF data',
      w2lapp.cnf.misc.ldif_maxbytes,1,ldif.ldif_pattern
    )
    self._max_entries = max_entries
    self.allRecords = []

  def getLDIFRecords(self):
    if self.value:
      f = StringIO('\n'.join(self.value).encode(self.charset))
      ldif_parser = ldif.LDIFRecordList(
        f,
        ignored_attr_types=[],
        max_entries=self._max_entries,
        process_url_schemes=w2lapp.cnf.misc.ldif_url_schemes
      )
      # It's not clear why ldif_parser stays persistent.
      # But let's reset class attribute now.
      # FIX ME!!! Likely not thread-safe!
      ldif_parser.all_records = []
      ldif_parser.parse()
      return ldif_parser.all_records[:]
    else:
      return []


class OIDInput(pyweblib.forms.Input):

  def __init__(self,name,text):
    pyweblib.forms.Input.__init__(
      self,name,text,
      512,1,u'[a-zA-Z0-9_.-]+',required=0,size=25
    )


class ObjectClassSelect(pyweblib.forms.Select):
  """Select field class for choosing the object class(es)"""
  def __init__(
    self,
    name='ldap_oc',
    text='Object classes',
    options=[],
    default=[],
    required=0,
    accesskey='',
    size=12,            # Size of displayed select field
  ):
    pyweblib.forms.Select.__init__(
      self,
      name,text,maxValues=200,required=0,
      options=options,default=default,
      accessKey=accesskey,
      size=size,
      ignoreCase=1,
      multiSelect=1
    )
    self.setRegex('[\w]+')
    self.maxLen = 200
  

class ExportFormatSelect(pyweblib.forms.Select):
  """Select field class for choosing export format"""

  def __init__(
    self,
    name='search_output',
    text=u'Export format',
    options=[],
    default='ldif1',
    required=0,
    accesskey='',
  ):
    pyweblib.forms.Select.__init__(
      self,
      name,text,1,
      options=[
        (u'table',u'Table/template'),
        (u'raw',u'Raw DN list'),
        (u'print',u'Printable'),
        (u'ldif',u'LDIF (Umich)'),
        (u'ldif1',u'LDIFv1 (RFC2849)'),
        (u'dsml1',u'DSMLv1'),
      ],
      default=default,
      required=required,
#      accesskey=accesskey,
      size=1,
    )


class AttributeType(pyweblib.forms.Input):
 
  def __init__(self,name,text,maxValues):
    pyweblib.forms.Input.__init__(
      self,name,text,500,maxValues,
      ur'[0-9.\w_;-]+',required=0,size=30
    )

